# Copyright 2019 Sam Vervaeck
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap
import ast

from .scanner import *
from .ast import *
from .util import *

PRECEDENCE_TABLE = [
    (2, '<',  1),
    (2, '<=', 1),
    (2, '>',  1), 
    (2, '>=', 1), 
    (2, '!=', 1),
    (2, '==', 1),
    (2, '|',  2),
    (2, '&',  3),
    (2, '<<', 4),
    (2, '>>', 4),
    (2, '+',  5),
    (2, '-',  5),
    (2, '*',  6),
    (2, '/',  6),
    (2, '@',  6),
    (2, '//', 6),
    (2, '%',  6),
    (1, '-',  7),
    (1, '*',  7),
    (1, '~',  7),
    (2, '**', 8),
    (2, '|>', 9)
    ]

def is_right_assoc(name):
    return name == '**'

def get_operator_precedence(name, arity):
    try:
        return next(prec for (arity2, name2, prec) in PRECEDENCE_TABLE if name == name2 and arity == arity2)
    except StopIteration:
        return None

#  def get_end_token_type(tt):
#      if tt == JOIN_KEYWORD:
#          return ENDJOIN_KEYWORD
#      elif tt == FOR_KEYWORD:
#          return ENDFOR_KEYWORD

class ParseError(RuntimeError):

    def __init__(self, filename, start_pos, end_pos, expected, actual):
        super().__init__("{}:{}:{}: Expected {} but got '{}'".format(filename, start_pos.line, start_pos.column, enum_or(token_type_to_string(tt) for tt in expected), actual))
        self._start_pos = start_pos
        self._end_pos = end_pos
        self._expected = expected
        self._actual = actual

    @property
    def start_pos(self):
        return self._start_pos

    @property
    def end_pos(self):
        return self._end_pos

class Parser:

    def __init__(self, scanner):
        self._scanner = scanner
        self._token_stream = scanner.scan()
        self._token_buffer = []
        self._statement_stack = []

    def peek_token(self, count=1):
        while len(self._token_buffer) < count:
            t0 = next(self._token_stream)
            self._token_buffer.append(t0)
        return self._token_buffer[count-1]

    def get_token(self):
        if len(self._token_buffer) > 0:
            return self._token_buffer.pop(0)
        return next(self._token_stream)

    def parse_pattern(self):
        t0 = self.get_token()
        if t0.type != IDENTIFIER:
            self._raise_parse_error(t0, [IDENTIFIER])
        return VarPattern(t0.value)

    def parse_member_expression(self):
        e = self.parse_prim_expression()
        path = []
        while True:
            t0 = self.peek_token()
            if t0.type == DOT:
                self.get_token()
                t1 = self.get_token()
                if t1.type != IDENTIFIER:
                    self._raise_parse_error(t1, [IDENTIFIER])
                path.append(t1.value)
            else:
                break
        if len(path) == 0:
            return e
        else:
            return MemberExpression(e, path)


    def parse_func_args(self):
        first = True
        while True:
            t0 = self.peek_token()
            if t0.type == END_OF_FILE:
                self._raise_parse_error(t0, [CLOSE_PAREN])
            elif t0.type == CLOSE_PAREN:
                break
            else:
                if not first:
                    if t0.type != COMMA:
                        self._raise_parse_error(t0, [COMMA])
                    self.get_token()
                else:
                    first = False
                yield self.parse_expression()

    def parse_func_app(self, e):
        t0 = self.get_token()
        if t0.type != OPEN_PAREN:
            self._raise_parse_error(t0, [OPEN_PAREN])
        args = list(self.parse_func_args())
        t1 = self.get_token()
        if t1.type != CLOSE_PAREN:
            self._raise_parse_error(t1, [CLOSE_PAREN])
        return AppExpression(e, args)

    def parse_app_expression(self):
        e = self.parse_member_expression()
        t1 = self.peek_token()
        if t1.type == OPEN_PAREN:
            return self.parse_func_app(e)
        elif t1.type == OPEN_BRACKET:
            return self.parse_arr_expr(e)
        else:
            return e

    def parse_unary_expression(self):
        heap = []
        while True:
            t0 = self.peek_token()
            if t0.type == OPERATOR:
                t0_prec = get_opeator_info(t0.value, 1)
                heapq.heappush((t0_prec, t0))
            else:
                break
        e = self.parse_app_expression()
        while len(heap) > 0:
            e = AppExpression(VarRefExpression(heappop(heap)[1].value), [e])
        return e

    def parse_prim_expression(self):
        t0 = self.get_token()
        if t0.type == OPEN_PAREN:
            e = self.parse_expression()
            self._expect_token(CLOSE_PAREN)
            return e
        elif t0.type == STRING_LITERAL:
            return ConstExpression(t0.value)
        elif t0.type == INTEGER:
            return ConstExpression(t0.value)
        elif t0.type == IDENTIFIER:
            return VarRefExpression(t0.value)
        else:
            self._raise_parse_error(t0, [IDENTIFIER, STRING_LITERAL, INTEGER, OPEN_PAREN])

    def parse_binary_operators(self, lhs, min_prec):
        t0 = self.peek_token()
        while True:
            if t0.type != OPERATOR:
                break
            keep = t0
            keep_prec = get_operator_precedence(keep.value, 2)
            if keep_prec is None:
                break
            self.get_token()
            rhs = self.parse_unary_expression()
            t0 = self.peek_token()
            while True:
                if not t0.type == OPERATOR:
                    break
                t0_prec = get_operator_precedence(t0.value, 2)
                if not (t0_prec is not None and (t0_prec > keep_prec or (is_right_assoc(t0.value) and t0_prec == keep_prec))):
                    break
                rhs = self.parse_binary_operators(rhs, t0_prec)
                t0 = self.peek_token()
            lhs = AppExpression(VarRefExpression(keep.value), [lhs, rhs])
        return lhs

    def parse_expression(self):
        return self.parse_binary_operators(self.parse_app_expression(), 0)

    def parse_expression_block(self):
        self._expect_token(OPEN_EXPRESSION_BLOCK)
        e = self.parse_expression()
        self._expect_token(CLOSE_EXPRESSION_BLOCK)
        return ExpressionStatement(e)

    def _raise_parse_error(self, token, expected):
        raise ParseError(self._scanner.get_filename(), token.start_pos, token.end_pos, expected, self._get_text(token))

    def _get_text(self, token):
        return token.get_text(self._scanner._data)

    def _expect_token(self, tt):
        t0 = self.get_token()
        if t0.type != tt:
            self._raise_parse_error(t0, [tt])

    def parse_statement(self):
        self._expect_token(OPEN_STATEMENT_BLOCK)
        t1 = self.get_token()
        if t1.type == IF_KEYWORD:
            self._statement_stack.append([ENDIF_KEYWORD, ELIF_KEYWORD, ELSE_KEYWORD])
            cond = self.parse_expression()
            self._expect_token(CLOSE_STATEMENT_BLOCK)
            then = list(self.parse_statement_block())
            cases = [(cond, then)]
            alt = None
            while True:
                self._expect_token(OPEN_STATEMENT_BLOCK)
                t2 = self.get_token()
                if t2.type == ELIF_KEYWORD:
                    cond = self.parse_expression()
                    self._expect_token(CLOSE_STATEMENT_BLOCK)
                    then = list(self.parse_statement_block())
                    cases.append((cond, then))
                elif t2.type == ELSE_KEYWORD:
                    self._expect_token(CLOSE_STATEMENT_BLOCK)
                    self._statement_stack[-1] = [ENDIF_KEYWORD]
                    alt = list(self.parse_statement_block())
                    self._expect_token(OPEN_STATEMENT_BLOCK)
                    self._expect_token(ENDIF_KEYWORD)
                    self._expect_token(CLOSE_STATEMENT_BLOCK)
                    break
                elif t2.type == ENDIF_KEYWORD:
                    self._expect_token(CLOSE_STATEMENT_BLOCK)
                    break
            return IfStatement(cases, alt)
        elif t1.type == FOR_KEYWORD:
            self._statement_stack.append([ENDFOR_KEYWORD])
            patt = self.parse_pattern()
            self._expect_token(IN_KEYWORD)
            e = self.parse_expression()
            self._expect_token(CLOSE_STATEMENT_BLOCK)
            body = list(self.parse_statement_block())
            self._expect_token(OPEN_STATEMENT_BLOCK)
            self._expect_token(ENDFOR_KEYWORD)
            self._expect_token(CLOSE_STATEMENT_BLOCK)
            return ForInStatement(patt, e, body)
        elif t1.type == JOIN_KEYWORD:
            self._statement_stack.append([ENDJOIN_KEYWORD])
            patt = self.parse_pattern()
            self._expect_token(IN_KEYWORD)
            e = self.parse_expression()
            self._expect_token(WITH_KEYWORD)
            sep = self.parse_expression()
            self._expect_token(CLOSE_STATEMENT_BLOCK)
            body = list(self.parse_statement_block())
            self._expect_token(OPEN_STATEMENT_BLOCK)
            self._expect_token(ENDJOIN_KEYWORD)
            self._expect_token(CLOSE_STATEMENT_BLOCK)
            return JoinStatement(patt, e, sep, body)
        elif t1.type == SETINDENT_KEYWORD:
            self._statement_stack.append([ENDSETINDENT_KEYWORD])
            e = self.parse_expression()
            self._expect_token(CLOSE_STATEMENT_BLOCK)
            body = list(self.parse_statement_block())
            self._expect_token(OPEN_STATEMENT_BLOCK)
            self._expect_token(ENDSETINDENT_KEYWORD)
            self._expect_token(CLOSE_STATEMENT_BLOCK)
            return SetIndentStatement(e, body)
        elif t1.type == NOINDENT_KEYWORD:
            self._statement_stack.append([ENDNOINDENT_KEYWORD])
            self._expect_token(CLOSE_STATEMENT_BLOCK)
            body = list(self.parse_statement_block())
            self._expect_token(OPEN_STATEMENT_BLOCK)
            self._expect_token(ENDNOINDENT_KEYWORD)
            self._expect_token(CLOSE_STATEMENT_BLOCK)
            return SetIndentStatement(ConstExpression(0), body)
        else:
            expected = [FOR_KEYWORD, JOIN_KEYWORD, IF_KEYWORD, NOINDENT_KEYWORD, SETINDENT_KEYWORD]
            if len(self._statement_stack) > 0:
                expected.extend(self._statement_stack[-1])
            self._raise_parse_error(t1, expected)

    def parse_statement_block(self):
        close_tts = self._statement_stack[-1]
        while True:
            t0 = self.peek_token(1)
            t1 = self.peek_token(2)
            if t0.type == OPEN_STATEMENT_BLOCK and t1.type in close_tts:
                break
            else:
                yield self.parse()

    def parse_code_block(self):
        stmts = []
        self._expect_token(OPEN_CODE_BLOCK)
        t0 = self.get_token()
        if t0.type != CODE_BLOCK_CONTENT:
            self._raise_parse_error(t0, [CODE_BLOCK_CONTENT])
        module = ast.parse(textwrap.dedent(t0.value))
        self._expect_token(CLOSE_CODE_BLOCK)
        return CodeBlock(module)

    def parse(self):
        t0 = self.peek_token()
        if t0.type == TEXT:
            self.get_token()
            return TextStatement(self._get_text(t0))
        elif t0.type == OPEN_CODE_BLOCK:
            return self.parse_code_block()
        elif t0.type == OPEN_EXPRESSION_BLOCK:
            return self.parse_expression_block()
        elif t0.type == OPEN_STATEMENT_BLOCK:
            return self.parse_statement()
        else:
            self._raise_parse_error(t0, [TEXT, OPEN_EXPRESSION_BLOCK, OPEN_STATEMENT_BLOCK])

    def parse_all(self):
        while True:
            t0 = self.peek_token()
            if t0.type == END_OF_FILE:
                break
            else:
                yield self.parse()




# Copyright 2019 Sam Vervaeck
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from .util import escape

def is_space(ch):
    return ch == '\t' or ch == '\n' or ch == '\r' or ch == ' '

def is_id_start(ch):
    return ch != EOF and re.match(r'^[a-zA-Z_]$', ch) is not None

def is_id_part(ch):
    return ch != EOF and re.match(r'^[a-zA-Z0-9_]$', ch) is not None

def is_digit(ch):
    return ch != EOF and re.match(r'^[0-9]$', ch) is not None

def pretty_char(ch):
    if ch == EOF:
        return 'end-of-file'
    else:
        return f'{escape(ch)}'

EOF = -1

TEXT                              = 0
IDENTIFIER                        = 1
STRING_LITERAL                    = 3
BOOLEAN                           = 4
IN_KEYWORD                        = 5
FOR_KEYWORD                       = 6
WHILE_KEYWORD                     = 7
ENDFOR_KEYWORD                    = 8
ENDWHILE_KEYWORD                  = 9
OPEN_EXPRESSION_BLOCK             = 10
CLOSE_EXPRESSION_BLOCK            = 11
OPEN_STATEMENT_BLOCK              = 12
CLOSE_STATEMENT_BLOCK             = 13
END_OF_FILE                       = 14
OPEN_PAREN                        = 15
CLOSE_PAREN                       = 16
OPERATOR                          = 17
OPEN_BRACKET                      = 18
CLOSE_BRACKET                     = 19
INTEGER                           = 20
COMMA                             = 21
JOIN_KEYWORD                      = 22
ENDJOIN_KEYWORD                   = 23
WITH_KEYWORD                      = 24
IF_KEYWORD                        = 25
ELSE_KEYWORD                      = 26
ENDIF_KEYWORD                     = 27
ELIF_KEYWORD                      = 28
OPEN_CODE_BLOCK                   = 29
CLOSE_CODE_BLOCK                  = 30
DOT                               = 31
NOINDENT_KEYWORD                  = 32
SETINDENT_KEYWORD                    = 33
DEDENT_KEYWORD                    = 34
ENDNOINDENT_KEYWORD               = 35
ENDSETINDENT_KEYWORD                 = 36
ENDDEDENT_KEYWORD                 = 37
CODE_BLOCK_CONTENT                = 38

OPERATORS = ['+', '-', '*', '**', '/', '//', '%', '@', '<<', '>>', '&', '|', '^', '~', ':=', '<', '>', '<=', '>=', '==', '!=', '|>']

def is_operator_start(ch):
    for op in OPERATORS:
        if op[0] == ch:
            return True
    return False

def is_operator_part(ch):
    for op in OPERATORS:
        if ch in op[1:]:
            return True
    return False

KEYWORDS = { 
    'for': FOR_KEYWORD, 
    'in': IN_KEYWORD, 
    'while': WHILE_KEYWORD, 
    'endfor': ENDFOR_KEYWORD, 
    'endwhile': ENDWHILE_KEYWORD, 
    'join': JOIN_KEYWORD, 
    'endjoin': ENDJOIN_KEYWORD,
    'with': WITH_KEYWORD,
    'if': IF_KEYWORD,
    'else': ELSE_KEYWORD,
    'endif': ENDIF_KEYWORD,
    'elif': ELIF_KEYWORD,
    'noindent': NOINDENT_KEYWORD,
    'endnoindent': ENDNOINDENT_KEYWORD,
    'setindent': SETINDENT_KEYWORD,
    'endsetindent': ENDSETINDENT_KEYWORD,
    'dedent': DEDENT_KEYWORD,
    'enddedent': ENDDEDENT_KEYWORD,
    }

TEXT_MODE = 0
STATEMENT_MODE = 1
CODE_BLOCK_MODE = 2

def token_type_to_string(tt):
    if tt == IDENTIFIER:
        return 'an identifier'
    elif tt == FOR_KEYWORD:
        return "'for'"
    elif tt == IN_KEYWORD:
        return "'in'"
    elif tt == ENDFOR_KEYWORD:
        return "'endfor'"
    elif tt == WHILE_KEYWORD:
        return "'while'"
    elif tt == ENDWHILE_KEYWORD:
        return "'while'"
    elif tt == STRING_LITERAL:
        return 'a string literal'
    elif tt == BOOLEAN:
        return "'True' or 'False'"
    elif tt == OPEN_EXPRESSION_BLOCK:
        return "'{{'"
    elif tt == CLOSE_EXPRESSION_BLOCK:
        return "'}}'"
    elif tt == OPEN_STATEMENT_BLOCK:
        return "'{%'"
    elif tt == CLOSE_STATEMENT_BLOCK:
        return "'%}'"
    elif tt == TEXT:
        return "some text"
    elif tt == OPEN_PAREN:
        return "'('"
    elif tt == CLOSE_PAREN:
        return "')'"
    elif tt == OPEN_BRACKET:
        return "'['"
    elif tt == CLOSE_BRACKET:
        return "']'"
    elif tt == INTEGER:
        return 'an integer'
    elif tt == JOIN_KEYWORD:
        return "'join'"
    elif tt == ENDJOIN_KEYWORD:
        return "'endjoin'"
    elif tt == WITH_KEYWORD:
        return "'with'"
    elif tt == IF_KEYWORD:
        return "'if'"
    elif tt == ELSE_KEYWORD:
        return "'else'"
    elif tt == ENDIF_KEYWORD:
        return "'endif'"
    elif tt == NOINDENT_KEYWORD:
        return "'noindent'"
    elif tt == SETINDENT_KEYWORD:
        return "'setindent'"
    elif tt == ENDNOINDENT_KEYWORD:
        return "'endnoindent'"
    elif tt == ENDSETINDENT_KEYWORD:
        return "'endsetindent'"

class Position:

    def __init__(self, offset=0, line=1, column=1):
        self.offset = offset
        self.line = line
        self.column = column

    def clone(self):
        return Position(self.offset, self.line, self.column)

    def advance(self, text):
        for ch in text:
            if ch == '\n':
                self.line += 1
                self.column = 0
            else:
                self.column += 1
            self.offset += 1

class Token:

    def __init__(self, type, start_pos, end_pos, value=None):
        self._type = type
        self._start_pos = start_pos
        self._end_pos = end_pos
        self._value = value

    def get_text(self, data):
        return data[self._start_pos.offset:self._end_pos.offset]

    @property
    def type(self):
        return self._type

    @property
    def value(self):
        return self._value

    @property
    def start_pos(self):
        return self._start_pos

    @property
    def end_pos(self):
        return self._end_pos

class ScanError(RuntimeError):

    def __init__(self, filename, start_pos, c0):
        super().__init__("{}:{}:{}: Got an unexpected {}".format(filename, start_pos.line, start_pos.column, pretty_char(c0)))
        self._start_pos = start_pos

    @property
    def start_pos(self):
        return self._start_pos

class Scanner:

    def __init__(self, filename, data, is_code=False):
        self._buffer = []
        self._data = data
        self._offset = 0
        self._filename = filename
        self._curr_pos = Position()
        self._mode = STATEMENT_MODE if is_code else TEXT_MODE 

    def get_filename(self):
        return self._filename

    def read_char(self):
        if self._offset == len(self._data):
            return EOF
        ch = self._data[self._offset]
        self._offset += 1
        return ch

    def peek_char(self, offset=1):
        while len(self._buffer) < offset:
            c0 = self.read_char()
            if c0 == EOF:
                return EOF
            self._buffer.append(c0)
        return self._buffer[offset-1]

    def get_char(self):
        if len(self._buffer) == 0:
            ch = self.read_char()
        else:
            ch = self._buffer.pop(0)
        if ch == '\n':
            self._curr_pos.line += 1
            self._curr_pos.column = 0
        else:
            self._curr_pos.column += 1
        self._curr_pos.offset += 1
        return ch

    def scan_raw_identifier(self):
        c0 = self.get_char()
        if not is_id_start(c0):
            raise ScanError(self._filename, self._curr_pos.clone(), c0)
        name = c0
        while True:
            ch1 = self.peek_char()
            if is_id_part(ch1):
                self.get_char()
                name += ch1
            else:
                break
        return name

    def skip_ws(self):
        while is_space(self.peek_char()):
            self.get_char()

    def scan_string_lit(self):
        value = ''
        start_pos = self._curr_pos.clone()
        escaping = False
        c0 = self.get_char()
        if c0 != '\'':
            raise ScanError(self._filename, self._curr_pos.clone(), c0)
        while True:
            ch = self.get_char()
            if ch == '\'':
                break
            elif ch == '\\':
                escaping = True
                continue
            if escaping:
                value += unescape(ch)
            else:
                value += ch
        return Token(STRING_LITERAL, start_pos, self._curr_pos.clone(), value)

    def scan_raw_text(self):
        text = ''
        start_pos = self._curr_pos.clone()
        while True:
            ch0 = self.peek_char()
            if ch0 == EOF:
                if len(text) > 0:
                    return Token(TEXT, start_pos, self._curr_pos.clone(), text) 
                return Token(END_OF_FILE, self._curr_pos.clone(), self._curr_pos.clone())
            elif ch0 == '{':
                if len(text) > 0:
                    return Token(TEXT, start_pos, self._curr_pos.clone(), text)
                self.get_char()
                ch1 = self.peek_char()
                if ch1 == '{':
                    self._mode = STATEMENT_MODE
                    self.get_char()
                    return Token(OPEN_EXPRESSION_BLOCK, start_pos, self._curr_pos.clone())
                elif ch1 == '%':
                    self._mode = STATEMENT_MODE
                    self.get_char()
                    return Token(OPEN_STATEMENT_BLOCK, start_pos, self._curr_pos.clone())
                elif ch1 == '!':
                    self._mode = CODE_BLOCK_MODE
                    self.get_char()
                    return Token(OPEN_CODE_BLOCK, start_pos, self._curr_pos.clone())
                else:
                    text += ch0
            else:
                self.get_char()
                text += ch0

    def scan(self):
        while True:
            if self._mode == TEXT_MODE:
                yield self.scan_raw_text()
            elif self._mode == CODE_BLOCK_MODE:
                in_string_literal = False
                start_pos = self._curr_pos.clone()
                text = ''
                while True:
                    c0 = self.peek_char(1)
                    c1 = self.peek_char(2)
                    if c0 == '!' and c1 == '}':
                        break
                    if c0 == EOF:
                        raise ScanError(self._filename, start_pos, c0)
                    text += self.get_char()
                self._mode = TEXT_MODE
                end_pos = self._curr_pos.clone()
                yield Token(CODE_BLOCK_CONTENT, start_pos, end_pos, text)
                start_pos = self._curr_pos.clone()
                self.get_char()
                self.get_char()
                end_pos = self._curr_pos.clone()
                yield Token(CLOSE_CODE_BLOCK, start_pos, end_pos)
            elif self._mode == STATEMENT_MODE:
                self.skip_ws()
                start_pos = self._curr_pos.clone()
                c0 = self.peek_char()
                if c0 == EOF:
                    yield Token(END_OF_FILE, self._curr_pos.clone(), self._curr_pos.clone())
                elif c0 == '.':
                    self.get_char()
                    yield Token(DOT, start_pos, self._curr_pos.clone())
                elif c0 == '!':
                    self.get_char()
                    c1 = self.get_char()
                    if c1 == '=':
                        yield Token(OPERATOR, start_pos, self._curr_pos.clone())
                    elif c1 == '}':
                        yield Token(CLOSE_CODE_BLOCK, start_pos, self._curr_pos.clone())
                    else:
                        raise ScanError(self._filename, self._curr_pos.clone(), c0)
                elif c0 == '%':
                    self.get_char()
                    c1 = self.get_char()
                    if c1 == '}':
                        self._mode = TEXT_MODE
                        yield Token(CLOSE_STATEMENT_BLOCK, start_pos, self._curr_pos.clone())
                    else:
                        yield Token(OPERATOR, start_pos, self._curr_pos.clone(), '%')
                elif c0 == '}':
                    self.get_char()
                    c1 = self.get_char()
                    if c1 == '}':
                        self._mode = TEXT_MODE
                        yield Token(CLOSE_EXPRESSION_BLOCK, start_pos, self._curr_pos.clone())
                    else:
                        raise ScanError(self._filename, self._curr_pos.clone(), c0)
                elif c0 == ',':
                    self.get_char()
                    yield Token(COMMA, start_pos, self._curr_pos.clone())
                elif c0 == '(':
                    self.get_char()
                    yield Token(OPEN_PAREN, start_pos, self._curr_pos.clone())
                elif c0 == ')':
                    self.get_char()
                    yield Token(CLOSE_PAREN, start_pos, self._curr_pos.clone())
                elif c0 == '[':
                    self.get_char()
                    yield Token(OPEN_BRACKET, start_pos, self._curr_pos.clone())
                elif c0 == ']':
                    self.get_char()
                    yield Token(CLOSE_BRACKET, start_pos, self._curr_pos.clone())
                elif c0 == '\'':
                    yield self.scan_string_lit()
                elif is_digit(c0):
                    self.get_char()
                    digits = c0
                    while is_digit(self.peek_char()):
                        digits += self.get_char()
                    yield Token(INTEGER, start_pos, self._curr_pos.clone(), int(digits))
                elif is_operator_start(c0):
                    op = c0
                    self.get_char()
                    while is_operator_part(self.peek_char()): 
                        op += self.get_char()
                    if not op in OPERATORS:
                        raise ScanError(self._filename, start_pos, op)
                    yield Token(OPERATOR, start_pos, self._curr_pos.clone(), op)
                elif is_id_start(c0):
                    name = self.scan_raw_identifier()
                    if name in KEYWORDS:
                        yield Token(KEYWORDS[name], start_pos, self._curr_pos.clone(), name)
                    else:
                        yield Token(IDENTIFIER, start_pos, self._curr_pos.clone(), name)
                else:
                    raise ScanError(self._filename, self._curr_pos.clone(), c0)




# Copyright 2019 Sam Vervaeck
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import ast

from .scanner import Position
from .ast import *
from .util import get_indentation, preorder, is_blank, starts_with_newline, ends_with_newline, is_empty, escape

from .lines import *

class Env:

    def __init__(self, parent=None):
        self.parent = parent
        self._variables = {}

    def lookup(self, name):
        if name in self._variables:
            return self._variables[name]
        if self.parent is not None:
            return self.parent.lookup(name)
        raise RuntimeError("Could not find '{}' in environment.".format(name))

    def set(self, name, value):
        self._variables[name] = value

    def fork(self):
        return Env(self)

def to_snake_case(name):
    if '-' in name:
        return name.replace('-', '_')
    else:
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

DEFAULT_BUILTINS = {
        'range': lambda a, b: range(a, b),
        '+': lambda a, b: a + b,
        '-': lambda a, b: a - b,
        '*': lambda a, b: a * b,
        '/': lambda a, b: a / b,
        '%': lambda a, b: a % b,
        '==': lambda a, b: a == b,
        'snake': to_snake_case,
        'upper': lambda s: s.upper(),
        'lower': lambda s: s.lower(),
        '|>': lambda val, f: f(val)
        }

def is_inner_wrapped(stmt):
    if not hasattr(stmt, 'body'):
        return False
    return len(stmt.body) > 0 \
       and isinstance(stmt.body[0], TextStatement) \
       and starts_with_newline(stmt.body[0].text) \
       and isinstance(stmt.body[-1], TextStatement) \
       and ends_with_newline(stmt.body[-1].text)

def get_inner_indentation(node, at_blank_line=True):
    curr_indent = 0
    min_indent = None
    for node in preorder(node):
        if isinstance(node, TextStatement):
            for ch in node.text:
                if not at_blank_line:
                    if ch == '\n':
                        at_blank_line = True
                        curr_indent = 0
                else:
                    if is_blank(ch):
                        curr_indent += 1
                        continue
                    at_blank_line = ch == '\n'
                    if not at_blank_line and (min_indent is None or curr_indent < min_indent):
                        min_indent = curr_indent
            if min_indent is None:
                min_indent = curr_indent
            return min_indent
        else:
            at_blank_line = False

def evaluate(ast, ctx={}, indentation='  ', filename="#<anonymous>"):

    if isinstance(ast, str):
        from .scanner import Scanner
        from .parser import Parser
        sc = Scanner(filename, ast)
        p = Parser(sc)
        ast = list(p.parse_all())

    curr_indent = ''
    at_blank_line = True
    out = Lines()

    def eval_code_expr(e, env):
        if isinstance(e, ConstExpression):
            return e.value
        elif isinstance(e, MemberExpression):
            out = eval_code_expr(e.expression, env)
            for name in e.path:
                out = out[name]
            return out
        elif isinstance(e, VarRefExpression):
            return env.lookup(e.name)
        elif isinstance(e, AppExpression):
            op = eval_code_expr(e.operator, env)
            args = list(eval_code_expr(arg, env) for arg in e.operands)
            if not callable(op):
                raise RuntimeError("Could not evaluate Templately expression: result is not applicable.".format(op))
            return op(*args)
        else:
            raise RuntimeError("Could not evaluate Templately expression: unknown expression {}.".format(e))

    def eval_statement_list(stmts, env):
        result = Lines()
        i = 0
        while i < len(stmts):
            stmt = stmts[i]
            last_indent = len(curr_indent)
            prev_line_blank = at_blank_line
            next_line_blank = i == len(stmts)-1 or (isinstance(stmts[i+1], TextStatement) and starts_with_newline(stmts[i+1].text))
            iter_result = eval_statement(stmt, env)
            if not isinstance(stmt, TextStatement) \
                    and prev_line_blank \
                    and next_line_blank:
                if last_indent > 0:
                    del result[-last_indent:]
                if len(iter_result) == 0:
                    del result[-1:]
            result += iter_result
            i += 1
        return result

    def indent_depth(indentation):
        return len(indentation)

    def update_locals(text):
        nonlocal at_blank_line, curr_indent
        last_indent = ''
        has_newline = False
        for ch in reversed(text):
            if ch == ' ':
                last_indent += ' '
            elif ch == '\n':
                has_newline = True
                at_blank_line = True
                break
            else:
                last_indent = ''
                at_blank_line = False
        if not has_newline:
            if at_blank_line:
                curr_indent = curr_indent + last_indent
        else:
            curr_indent = last_indent

    def eval_repeat(stmt, sep, env):
        result = Lines()
        rng = eval_code_expr(stmt.expression, env)
        count = max(rng) - min(rng) + 1
        env2 = env.fork()
        outer_indent = len(curr_indent)
        inner_indent = get_inner_indentation(stmt, at_blank_line)
        wrapped = is_inner_wrapped(stmt)
        for i in range(0, count):
            if i > 0: result += sep
            env2.set(stmt.pattern.name, i + min(rng))
            iter_result = eval_statement_list(stmt.body, env2)
            if wrapped and i > 0:
                # remove first newline for i > 0
                # rest will be removed by dedent()
                del iter_result[0:1]
            result += iter_result
        if wrapped:
            result.dedent()
            del result[0:1]
            del result[-1:]
        result.indent(' ' * outer_indent)
        return result


    def eval_statement(stmt, env):

        nonlocal curr_indent, at_blank_line, out

        if isinstance(stmt, TextStatement):
            update_locals(stmt.text)
            text = stmt.text
            result = split_lines(text)
            out += result 
            return result

        elif isinstance(stmt, IfStatement):
            for (cond, cons) in stmt.cases:
                if eval_code_expr(cond, env):
                    env2 = env.fork()
                    result = eval_statement_list(cons, env2)
                    out += result
                    return result
            if stmt.alternative is not None:
                env2 = env.fork()
                result = eval_statement_list(stmt.alternative, env2)
                out += result
                return result
            return Lines()

        elif isinstance(stmt, CodeBlock):
            exec(compile(stmt.module, filename=filename, mode='exec'), global_env._variables, env._variables)
            return Lines()

        elif isinstance(stmt, SetIndentStatement):
            outer_indent = len(curr_indent)
            inner_indent = get_inner_indentation(stmt)
            indent_override = eval_code_expr(stmt.level, env)
            result = eval_statement_list(stmt.body, env)
            if is_inner_wrapped(stmt):
                result.dedent()
                del result[0:1]
                del result[-1:]
            result.indent(' ' * outer_indent)
            for line in result.get_lines():
                if line.indent_override is None:
                    line.indent_override = indent_override
            out += result
            return result

        elif isinstance(stmt, ExpressionStatement):
            result = split_lines(str(eval_code_expr(stmt.expression, env)))
            out += result
            return result

        elif isinstance(stmt, ForInStatement):
            result = eval_repeat(stmt, Lines(), env)
            out += result
            return result

        elif isinstance(stmt, JoinStatement):
            sep = split_lines(str(eval_code_expr(stmt.separator, env)))
            result = eval_repeat(stmt, sep, env)
            out += result
            return result

        else:
            raise TypeError("Could not evaluate statement: unknown statement {}.".format(stmt))

    global_env = Env()
    global_env.set('True', True)
    global_env.set('False', False)
    global_env.set('now', datetime.now().strftime("%b %d %Y %H:%M:%S"))
    for name, value in DEFAULT_BUILTINS.items():
        global_env.set(name, value)
    for name, value in ctx.items():
        global_env.set(name, value)

    output = ''
    res = eval_statement_list(ast, global_env)
    lines = res._lines
    i = 0

    while i < len(lines):

        indent_override = lines[i].indent_override

        # look ahead to see if this line or the next may contain
        # an indent_override flag
        j = i + 1
        while j < len(lines) and lines[j].join_with_next:
            if lines[j].indent_override is not None:
                indent_override = lines[j].indent_override
                break
            j += 1

        # no indent_override means there is no
        # special processing that has to take place
        if indent_override is None:
            output += str(lines[i])

        else:

            curr_indent_override = indent_override
            k = 0

            while i < len(lines):

                # pass through any characters that
                # are accepted by curr_indent_override
                while k < len(lines[i].text):
                    ch = lines[i].text[k]
                    if curr_indent_override == 0:
                        break
                    elif is_blank(ch):
                        curr_indent_override -= 1
                        output += ch
                        k += 1
                    else:
                        output += ' ' * curr_indent_override
                        break

                # FIXME currently no code requires this
                # if the line was not long enough and the next line might still
                # contain indentation, then now is the time to process it
                #  if curr_indent_override > 0:
                #      i += 1
                #      continue

                # skip any characters that are remaining
                # because they are excess indentation
                while k < len(lines[i].text) and is_blank(lines[i].text[k]):
                    k += 1

                # now the real string is added
                output += lines[i].text[k:]

                # finish off unless we have to append more lines
                if not lines[i].join_with_next:
                    output += '\n'
                    break

                curr_indent_override = indent_override
                i += 1

        i += 1

    return output






from .main import main
from .evaluator import evaluate

def execute(filepath, ctx={}, **kwargs):
    with open(filepath, 'r') as f:
        contents = f.read()
    return evaluate(contents, ctx, filename=os.path.basename(filepath), **kwargs)




# Copyright 2019 Sam Vervaeck
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for working with lines of text.

These classes resemble the ordinary string operations, but they were
fine-tuned to work very efficiently on lists of special line objects.
"""


from .util import is_blank


class Line:


    def __init__(self, text, join_with_next=False, indent_override=None):
        self.indent_override = indent_override
        self.join_with_next = join_with_next
        self.text = text


    def __str__(self):
        out = self.text
        if not self.join_with_next:
            out += '\n'
        return out


    def clone(self):
        return Line(self.text, self.join_with_next, self.indent_override)


    def __getitem__(self, index):
        if isinstance(index, int):
            if index < len(self.text):
                return self.text[index]
            if index == len(self.text) and not self.join_with_next:
                return '\n'
            raise IndexError('index out of bounds')
        else:
            raise NotImplementedError("type is not supported")


    def __delitem__(self, item):
        if isinstance(item, int):
            if item < 0:
                item = item + len(self)
            if item < len(self.text):
                self.text = self.text[0:item] + self.text[0:item+1]
            if item == len(self.text):
                self.join_with_next = True
        elif isinstance(item, slice):
            start_offset = 0 if item.start is None else item.start
            end_offset = len(self) if item.stop is None else item.stop
            assert(0 <= start_offset < len(self))
            assert(0 <= end_offset <= len(self))
            assert(start_offset <= end_offset)
            if end_offset > len(self.text):
                self.join_with_next = True
            self.text = self.text[:start_offset] + self.text[end_offset:]
        else:
            raise TypeError("item must be a slice or an integer")


    def __len__(self):
        length = len(self.text)
        if not self.join_with_next:
            length += 1
        return length


    def prepend(self, prefix):
        self.text = prefix + self.text


class Lines:


    def __init__(self, lines=[]):
        self._lines = list(lines)


    def __str__(self):
        out = ''
        for line in self._lines:
            out += str(line)
        return out


    def get_lines(self):
        return iter(self._lines)


    def clone(self):
        return Lines(line.clone() for line in self._lines)


    def __iadd__(self, other):
        self._lines.extend(other._lines)
        return self


    def __add__(self, other):
        return Lines(self._lines + other._lines)


    def __len__(self):
        return sum(len(line) for line in self._lines)


    def get_line(self, index):
        return self._lines[index]


    def __getitem__(self, item):

        if isinstance(item, int):
            for line in self._lines:
                line_length = len(line)
                if item < line_length:
                    return line[item]
                else:
                    item -= line_length
            raise IndexError("lines index out of range")
        else:
            raise NotImplementedError("type is not supported")


    def __delitem__(self, item):

        if isinstance(item, int):
            raise NotImplementedError("deleting by index is not yet implemented")

        elif isinstance(item, slice):

            start_offset = item.start
            end_offset = item.stop

            if start_offset is None:
                start_offset = 0
            elif start_offset < 0:
                start_offset = len(self) + start_offset

            if end_offset is None:
                end_offset = len(self)
            elif end_offset < 0:
                end_offset = len(self) + end_offset

            i = 0
            j = start_offset
            total_to_remove = end_offset - start_offset

            if total_to_remove < 0:
                raise IndexError("start offset goes beyond end offset")

            while i < len(self._lines):
                line = self._lines[i]
                if j < len(line):
                    can_be_removed = len(line) - j
                    if j == 0 and total_to_remove >= can_be_removed:
                        del self._lines[i]
                        total_to_remove -= can_be_removed
                    else:
                        to_remove = min(total_to_remove, can_be_removed)
                        del line[j:j+to_remove]
                        total_to_remove -= to_remove
                        i += 1
                    break
                else:
                    j -= len(line)
                    i += 1

            while i < len(self._lines):
                line = self._lines[i]
                can_be_removed = len(line)
                if total_to_remove <= 0:
                    break 
                elif total_to_remove >= can_be_removed:
                    del self._lines[i]
                    total_to_remove -= can_be_removed
                else:
                    to_remove = min(total_to_remove, can_be_removed)
                    del line[0:to_remove]
                    total_to_remove -= to_remove
                    break

        else:
            raise TypeError(f"item {item} must be a slice or a number")


    def indent(self, indentation='  ', at_blank_line=True, join_with_prev=False):
        for line in self._lines:
            if at_blank_line:
                for ch in line.text:
                    if not is_blank(ch):
                        at_blank_line = False
                        line.prepend(indentation)
                        break
            if not line.join_with_next:
                at_blank_line = True

    def get_indentation(self, at_blank_line=True, default_indent=0):
        min_indent = None
        curr_indent = 0
        for line in self._lines:
            if at_blank_line:
                for ch in line.text:
                    if is_blank(ch):
                        curr_indent += 1
                    elif min_indent is None or curr_indent < min_indent:
                        min_indent = curr_indent
            if not line.join_with_next:
                at_blank_line = True
                curr_indent = 0
        if min_indent is None:
            min_indent = default_indent
        return min_indent

    def dedent(self, at_blank_line=True):
        min_indent = self.get_indentation(at_blank_line=at_blank_line)
        for line in self._lines:
            if at_blank_line:
                i = 0
                while i < len(line.text) and i < min_indent: 
                    if not is_blank(line.text[i]):
                        at_blank_line = False
                        break
                    i += 1
                if len(line.text) > 0:
                    del line[0:i]
            elif not line.join_with_next:
                at_blank_line = True


def split_lines(text):
    lines = []
    buffered = ''
    for ch in text:
        if ch == '\n':
            lines.append(Line(buffered))
            buffered = ''
        else:
            buffered += ch
    if len(buffered) > 0:
        lines.append(Line(buffered, True))
    return Lines(lines)




# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='templaty',
    version='1.3.0',
    description='Generate programming code from template files',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/samvv/Templaty',
    author='Sam Vervaeck',
    author_email='vervaeck.sam@skynet.be',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Code Generators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='template code-generator c++ c php java python',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    python_requires='>=3.5, <4',
    install_requires=['gast'],
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
    entry_points={
        'console_scripts': [
            'templaty=templaty:main',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/samvv/Templaty/issues',
        'Source': 'https://github.com/samvv/Templaty/',
    },
)




import os, sys
import argparse
from importlib import import_module
import json

from .ast import set_all_parents
from .scanner import Scanner
from .parser import Parser
from .evaluator import evaluate

def main(argv=None):

    parser = argparse.ArgumentParser()
    parser.add_argument('file', help='The template file from which code will be generated.')
    input_flags = parser.add_mutually_exclusive_group()
    input_flags.add_argument('--data-file', help='A JSON file containing variables that will be passed to the template')
    input_flags.add_argument('--stdin', action='store_true', help='When present, reads JSON data from STDIN and passes it to the template')

    args = parser.parse_args(argv)

    if args.data_file is not None:
        with open(args.data_file, 'r') as f:
            data = json.loads(f.read())
    elif args.stdin:
        data = json.loads(sys.stdin.read())
    else:
        data = {}

    with open(args.file, 'r') as f:
        contents = f.read()

    sc = Scanner(args.file, contents)
    p = Parser(sc)
    ss = list(p.parse_all())
    for s in ss:
        set_all_parents(s)

    print(evaluate(ss, data, filename=args.file))





# Copyright 2019 Sam Vervaeck
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
from typing import List, Optional, Any, Tuple, Union

from .util import BaseNode

class Node(BaseNode):
    pass

class Pattern(Node):
    pass

class VarPattern(Pattern):
    name: str

class Expression(Node):
    pass

class MemberExpression(Expression):
    expression: Expression
    path: List[str]

class VarRefExpression(Expression):
    name: str

class ConstExpression(Expression):
    value: Any

class AppExpression(Expression):
    operator: Expression
    operands: List[Expression]

class Statement(Node):
    pass

class TextStatement(Statement):
    text: str

class ExpressionStatement(Statement):
    expression: Expression

class IfStatement(Statement):
    cases: List[Tuple[Expression, List[Statement]]]
    alternative: Optional[List[Statement]]

class CodeBlock(Statement):
    module: ast.Module

class SetIndentStatement(Statement):
    level: int
    body: List[Statement]

class JoinStatement(Statement):
    pattern: Pattern
    expression: Expression
    separator: Expression
    body: List[Statement]

class ForInStatement(Statement):
    pattern: Pattern
    expression: Expression
    body: List[Statement]

def set_all_parents(node, parent=None):
    node.parent = parent
    for child in node.get_child_nodes():
        set_all_parents(child, node)




# Copyright 2019 Sam Vervaeck
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

def enum_or(raw_elements):
    elements = list(raw_elements)
    if len(elements) == 1:
        return str(elements[0])
    else:
        return ', '.join(str(el) for el in elements[0:-1]) + ' or ' + str(elements[-1])

def is_blank(text):
    for ch in text:
        if not ch == ' ' and not ch == '\t':
            return False
    return True

def is_empty(text):
    for ch in text:
        if not is_blank(ch) and not ch == '\n':
            return False
    return True

def starts_with_newline(text):
    for ch in text:
        if ch == '\n':
            return True
        if ch == ' ' or ch == '\t' or ch == '\r':
            continue
        break
    return False

def ends_with_newline(text):
    for ch in reversed(text):
        if ch == '\n':
            return True
        if ch == ' ' or ch == '\t' or ch == '\r':
            continue
        break
    return False

def remove_last_newline(text):
    i = 0
    while i < len(text):
        ch = text[len(text)-i]
        if ch == '\n':
            i -= 1
            break
        if not is_blank(ch):
            return text
        i -= 1
    while i < len(text) and is_blank(text[len(text)-i]):
        i += 1
    return text[0:i+1]

def dedent(text, at_blank_line=True):
    indent_length = get_indentation(text, at_blank_line)
    curr_removed = 0
    out = ''
    for ch in text:
        if ch == '\n':
            at_blank_line = True
        elif is_blank(ch):
            if curr_removed < indent_length:
                curr_removed += 1
                continue
        else:
            curr_removed = 0
        out += ch
    return out

def indent(text, indentation='  ', at_blank_line=True):
    out = ''
    for ch in text:
        if not at_blank_line:
            at_blank_line = ch == '\n'
        elif ch == '\n':
            at_blank_line = True
        elif not is_blank(ch):
            at_blank_line = False
            out += indentation
        out += ch
    return out

def is_last_line_blank(text):
    for ch in reversed(text):
        if ch == '\n':
            break
        elif not is_blank(ch):
            return False
    return True

def get_indentation(text, at_blank_line=True, default_indent=0):
    min_indent = None
    curr_indent = 0
    for ch in text:
        if not at_blank_line:
            if ch == '\n':
                at_blank_line = True
                curr_indent = 0
        else:
            if is_blank(ch):
                curr_indent += 1
                continue
            at_blank_line = ch == '\n'
            if not at_blank_line and (min_indent is None or curr_indent < min_indent):
                min_indent = curr_indent
    if min_indent is None:
        min_indent = default_indent
    return min_indent

def to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def breadthfirst(root, expand):
    queue = [root]
    while len(queue) > 0:
        node = queue.pop(0)
        yield node
        for node in reversed(list(expand(node))):
            queue.append(node)

def preorder(root, expand):
    stack = [root]
    while len(stack) > 0:
        node = stack.pop()
        yield node
        for node in expand(node):
            stack.append(node)

def postorder(root, expand):
    stack_1 = [root]
    stack_2 = []
    while len(stack_1) > 0:
        node = stack_1.pop()
        stack_2.append(node)
        for child in reversed(list(expand(node))):
            stack_1.append(child)
    for node in reversed(stack_2):
        yield node

def set_all_parents(node, parent=None):
    node.parent = parent
    for child in gast.iter_child_nodes(node):
        set_all_parents(child, node)

SPECIAL_CHARS = {
        '\x09': '\\t',
        '\x0D': '\\r',
        '\x0A': '\\n',
        '\x5C': '\\',
        }

def escape(text):
    out = ''
    for ch in text:
        if ch in SPECIAL_CHARS:
            out += SPECIAL_CHARS[ch]
        elif ch.isprintable():
        #  elif ord(ch) >= 0x20 and ord(ch) <= 0x7E:
            out += ch
        elif ord(ch) <= 0x7F:
            out += "\\x{:02X}".format(ord(ch))
        else:
            out += '\\u{:04X}'.format(ord(ch))
    return out

import typing

def is_list_type(ty):
    return hasattr(ty, '__origin__') and ty.__origin__ == list

def is_union_type(ty):
    return hasattr(ty, '__origin__') and ty.__origin__ == typing.Union

def is_none_type(ty):
    return ty == type(None)

def flatten_union_type(ty):
    if is_union_type(ty):
        return ((el2_ty for el2_ty in flatten_union_type(el_ty)) for el_ty in ty.__args__)
    else:
        return [ty]

def is_type_optional(ty):
    return any(is_none_type(el_ty) for el_ty in flatten_union_type(ty))

def get_element_type(ty):
    if is_list_type(ty):
        return ty.__args__[0]
    elif is_union_type(ty):
        return next(get_element_type(el_ty) for el_ty in flatten_union_type(ty) if not is_none_type(el_ty))
    else:
        return ty

def is_special_type(ty):
    return ty == typing.Any or is_none_type(ty)

def get_index_type(ty):
    if is_list_type(ty):
        return int
    else:
        return None

def preorder(node):
    stack = [node]
    while len(stack) > 0:
        front = stack.pop()
        yield front
        for node in front.get_child_nodes():
            stack.append(node)

class BaseNode:

    def __init__(self, *args, **kwargs):
        self.parent = None
        fields = self.__dict__['_fields'] = dict()
        i = 0
        for name, ty in self.__class__.__annotations__.items():
            if name in kwargs:
                value = kwargs[name]
            else:
                value = args[i]
                i += 1
            fields[name] = value
        for name, ty in self.__class__.__annotations__.items():
            if not name in fields:
                if is_list_type(ty):
                    fields[name] = []
                elif is_type_optional(ty):
                    fields[name] = None
                else:
                    raise TypeError(f"field '{name}' is required but did not receive a value")

    def get_field_names(self):
        return iter(self.__class__.__annotations__.keys())

    def get_child_nodes(self):
        for name, value in self.__dict__['_fields'].items():
            if value is not None:
                ty = get_element_type(self.__class__.__annotations__[name])
                if not is_special_type(ty) and issubclass(ty, BaseNode):
                    if isinstance(value, list):
                        for el in value:
                            yield el
                    else:
                        yield value

    def get_field_items(self):
        for name, value in self.__dict__['_fields'].items():
            if value is not None:
                if isinstance(value, list):
                    for el in value:
                        yield (name, el)
                else:
                    yield (name, value)

    def __getattr__(self, name):
        fields = self.__dict__['_fields']
        if name in fields:
            return fields[name]
        if name in self.__dict__:
            return self.__dict__[name]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        if name in self.__class__.__annotations__:
            self.__dict__['_fields'][name] = value
        else:
            self.__dict__[name] = value

